
#define OSD_OPTIONS      (DLOP_ALPHACHANNEL | DLOP_OPACITY | DLOP_FLICKER_FILTERING) //  | DLOP_SRC_COLORKEY

DisplayLayerFuncs  oldPrimaryFuncs;
void              *oldPrimaryDriverData;

static CorePalette *private_pal;

/*******************************************************************************/

static void _cx2450x_osd_set_core_palette(CorePalette *pal)
{
    int i;

    if (!pal)
	return;

    if (pal->num_entries != 256)
	return;

    for (i = 0; i < 256; i++)
    {
	pal->entries[i].a = i;
	pal->entries[i].r = i;
	pal->entries[i].g = i;
	pal->entries[i].b = i;
    }
}

/*******************************************************************************/

static DFBResult osdInitLayer(CoreLayer                  *layer,
            		      void                       *driver_data,
			      void                       *layer_data,
			      DFBDisplayLayerDescription *description,
			      DFBDisplayLayerConfig      *config,
			      DFBColorAdjustment         *adjustment)
{
    DFBResult ret;

    /* call the original initialization function first */
    ret = oldPrimaryFuncs.InitLayer(layer, oldPrimaryDriverData, layer_data, description, config, adjustment);

    if (ret)
	return ret;

    /* set name */
    snprintf(description->name, DFB_DISPLAY_LAYER_DESC_NAME_LENGTH, "NXP CX2450x OSD Layer");

    /* add support for options and some defaults */
    config->flags       |= DLCONF_WIDTH | DLCONF_HEIGHT | DLCONF_OPTIONS | DLCONF_PIXELFORMAT;
    config->pixelformat  = DSPF_ARGB;
    config->width        = 1280;
    config->height       = 720;
    config->options      = DLOP_ALPHACHANNEL | DLOP_FLICKER_FILTERING;

    /* add some capabilities */
    description->caps = DLCAPS_SURFACE | DLCAPS_ALPHACHANNEL | DLCAPS_FLICKER_FILTERING | DLCAPS_OPACITY; //  | DLCAPS_SRC_COLORKEY
    description->type = DLTF_GRAPHICS;

    return DFB_OK;
}

/*******************************************************************************/

static DFBResult osdTestRegion(CoreLayer                  *layer,
            		       void                       *driver_data,
	                       void                       *layer_data,
			       CoreLayerRegionConfig      *config,
			       CoreLayerRegionConfigFlags *failed)
{
    DFBResult                  ret;
    CoreLayerRegionConfigFlags fail = 0;
    DFBDisplayLayerOptions     options = config->options;

    /* remove options before calling the original function */
    config->options = DLOP_NONE;

    /* call the original function */
    ret = oldPrimaryFuncs.TestRegion(layer, oldPrimaryDriverData, layer_data, config, &fail);

    /* check options if specified */
    if (options)
    {
	/* any unsupported option wanted? */
	if (options & ~OSD_OPTIONS)
	    fail |= CLRCF_OPTIONS;
    }

    /* restore options */
    config->options = options;

    if (failed)
	*failed = fail;

    if (fail)
	return DFB_UNSUPPORTED;

    return ret;
}

/*******************************************************************************/

static DFBResult osdSetRegion(CoreLayer                  *layer,
            		      void                       *driver_data,
	            	      void                       *layer_data,
			      void                       *region_data,
			      CoreLayerRegionConfig      *config,
			      CoreLayerRegionConfigFlags  updated,
			      CoreSurface                *surface,
			      CorePalette                *palette,
                              CoreSurfaceBufferLock      *left_lock,
                              CoreSurfaceBufferLock      *right_lock)
{
    DFBResult ret;
    CX2450XDriverData *drv = (CX2450XDriverData *)driver_data;
    volatile u32 *reg;
    CorePalette *orig_pal = NULL;
    FBDev *dfb_fbdev = dfb_system_data();
    CNXTFB_BLEND_MODE blend_mode;

    printf("%s() opacity: 0x%.8X, pixelformat: 0x%.8X, updated: 0x%.8X\n", __FUNCTION__, config->opacity, config->format, updated);

    if (!DFB_PIXELFORMAT_HAS_ALPHA(config->format))
	config->options &= ~DLOP_ALPHACHANNEL;	/* prevent enabling of alphachannel on formats without alpha */


    if (!DFB_PIXELFORMAT_IS_INDEXED(config->format))
    {
	/* the CX2450x handles all pixelformat with paelttes except for 24/32 bit YUV, so we need to create our own palette. 
	   Thy direct YUV formats of the CX2450x do not fit to DFB's formats (CX2450: AYCrCb 4444, AYCrCb 2644, YCrCb 655 
	   and AYCrCb 8888), so we ignore them here. */

	    orig_pal = palette;		/* save the pointer to original palette... if there is a palette */
	    palette = private_pal;

	    _cx2450x_osd_set_core_palette(palette);

	    /* call the original function */
	    ret = oldPrimaryFuncs.SetRegion(layer, oldPrimaryDriverData, layer_data, region_data, config, updated, surface, palette, left_lock, right_lock);

	    palette = orig_pal;		/* restore the original */
	    updated &= ~CLRCF_PALETTE;
    }
    else
    {
        /* call the original function without modifications */
	ret = oldPrimaryFuncs.SetRegion(layer, oldPrimaryDriverData, layer_data, region_data, config, updated, surface, palette, left_lock, right_lock);
    }

    if (ret)
	return ret;

    printf("options: \n");
    if (config->options == 0)
	printf("    0x0000000: DLOP_NONE\n");
    if (config->options & DLOP_ALPHACHANNEL)
	printf("    0x00000001: DLOP_ALPHACHANNEL\n");
    if (config->options & DLOP_FLICKER_FILTERING)
	printf("    0x00000002: DLOP_FLICKER_FILTERING\n");
    if (config->options & DLOP_DEINTERLACING)
	printf("    0x00000004: DLOP_DEINTERLACING\n");
    if (config->options & DLOP_SRC_COLORKEY)
	printf("    0x00000008: DLOP_SRC_COLORKEY\n");
    if (config->options & DLOP_DST_COLORKEY)
	printf("    0x00000010: DLOP_DST_COLORKEY\n");
    if (config->options & DLOP_OPACITY)
	printf("    0x00000020: DLOP_OPACITY\n");
    if (config->options & DLOP_FIELD_PARITY)
	printf("    0x00000040: DLOP_FIELD_PARITY\n");
    if (config->options & DLOP_LR_MONO)
	printf("    0x00000100: DLOP_LR_MONO\n");
    if (config->options & DLOP_STEREO)
	printf("    0x00000200: DLOP_STEREO\n");
    if ((config->options & DLOP_ALL) == DLOP_ALL)
	printf("    0x00003FF: DLOP_ALL\n");

    /* Setup opacity */
    if (config->options &  DLOP_OPACITY)
    {
        printf("--- global alpha ---\n");
        if (ioctl(dfb_fbdev->fd, FBIO_SETOPACITY, (u_int8_t)config->opacity)) {
             D_DEBUG_AT( cx2450x, "Can't set opacity -> %s\n", strerror( errno ) );
        }
        blend_mode = CNXTFB_BLEND_MODE_GLOBAL_ALPHA;
    }

    if (config->options & DLOP_ALPHACHANNEL) /* Pixel alpha */
    {
        printf("--- pixel alpha ---\n");
        blend_mode = CNXTFB_BLEND_MODE_PIXEL_ALPHA;
    }

    if ((config->options & DLOP_ALPHACHANNEL) && (config->options &  DLOP_OPACITY))	/* maybe DFB do this, we can both together by multiply */
    {
        printf("--- pixel alpha * opacity ---\n");
       /* enable multiply */
       blend_mode = CNXTFB_BLEND_MODE_ALPHA_MULTIPLIED;
    }

    if (ioctl(dfb_fbdev->fd, FBIO_SETBLENDMODE, (u_int8_t)blend_mode))
        D_DEBUG_AT( cx2450x, "Can't set blend mode -> %s\n", strerror( errno ) );

    /* Flicker filter */
    if (ioctl(dfb_fbdev->fd, FBIO_SETFLICKERFILTER, (config->options & DLOP_FLICKER_FILTERING) ? 1 : 0))
        D_DEBUG_AT( cx2450x, "Can't set flicker filter  -> %s\n", strerror( errno ) );

    return DFB_OK;
}

/*******************************************************************************/

DisplayLayerFuncs newPrimaryFuncs = 
{
    .InitLayer  = osdInitLayer,
    .TestRegion = osdTestRegion,
    .SetRegion  = osdSetRegion,
};

